/**
 * A class providing the storage for the simulation
 */

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.event.*;
import java.io.*;
import javax.swing.JOptionPane;
import javax.swing.JFileChooser;
import java.util.Date;
import java.util.StringTokenizer;
import java.util.LinkedList;

/**
 * an essentially static class containing the simulation
 */
public class Model {
    
   /**
   *  MODEL Section
   */
    
    /** Title of the main window  **/
    public static String title = "Bezier Patches";
    /** Labels on the custom buttons  **/
    public static String Labels[] = { "Reset","Pan", "Zoom", "Rotate", "Reshape"};
    /** repainting the screen */
    public boolean doRepaint;
    /** solution is running */
    public boolean isRunning;
    /** painting the screen to protect the data while painting */
    private boolean painting;
    /** system initialized? */
    private boolean initialized;
    /* time */
    private int t;
    
    private Color blink;
    private boolean isBlack;
    // mouse modes
    private static final int NONE = 0;
    private static final int PAN = 1;
    private static final int ZOOM = 2;
    private static final int ROTATE = 3;
    private static final int RESHAPE = 4;
    private int mode = PAN;
    private int baseX;
    private int baseY;
    private Vector baseOffset;
    private double baseScale;
    private double baseTheta;
    private double basePhi;
    private int blinkRate;
    private Vector pointedCoordinate;
    private Vector baseCtl;
    private Patch thisPatch;
    /**
     * basic constructor
     */
    public Model() {
        Global.doDbg = false;
        reset();
    }
    
    private void setSize(Dimension d) {
        Global.screenSize = d;
    }
    
    
    private void reset() {
        t = 0;
        Global.DT = 0.03333333;
        blinkRate = (int)(0.4/Global.DT);
        initialized = false;
        isBlack = true;
        blink = Color.BLACK;
        Global.offset = new Vector(Global.screenSize.width/2, 
                                    Global.screenSize.height/2);
        Global.scale = 2;
        Global.phi = -25.0 * Math.PI /  180; // radians
        Global.theta = 0.0; // radians
        mode = PAN;
    }
    
    
    /**
     * hit here each clock tick
     */
    public void timerTick() {
        t++;
        
        if(t%blinkRate == 0) {
          if(isBlack) {
            blink = Color.WHITE;
          } else {
            blink = Color.BLACK;
          }
          isBlack = !isBlack;
        }
        if(!painting) doRepaint = true;
    }
    
    
    /**
     *  VIEW Section
     */
    
    private Color backColor = new Color(150, 255, 150);
    
    
    private void initializeDisplay(Dimension d) {
        Global.screenSize = d;
        initialized = true;
        thisPatch = new Patch(0);
    }
    
    /**
     * paint the screen
     * @param g the graphic environment
     * @param d the screen size
     */
    public void doPaint(Graphics g, Dimension d) {
        
        setSize(d);
        if(!initialized) initializeDisplay(d);
        painting = true;
        int floor = d.height - 10;
        g.setColor(backColor);
        g.fillRect(0, 0, d.width, d.height);
        // Do all drawing here
        g.setColor(blink);
        g.fillRect(10, floor-20, 20, 20);
	Global.getEyeCoordinateAndTransformationMatrix();
	thisPatch.getCtrlPointsTransformatedCoordinate();
	thisPatch.getBezierPointsTransformatedCoordinate();
	thisPatch.getBezierSurfacesZvalue();
	thisPatch.processSortBezierSurfacesByZvalue();
	thisPatch.drawBezierSurfaces(g);
	thisPatch.drawCtrlPoints(g);
        g.setColor(Color.BLACK);
        g.drawString("If this box isn't blinking, press 'Start'", 40, floor);
        switch(mode) {
            case NONE:
                break;
            case PAN:
                g.drawString("Pan mode", d.width/2, floor);
                break;
            case ZOOM:
                g.drawString("Zoom mode", d.width/2, floor);
                break;
            case ROTATE:
                g.drawString("Rotate mode", d.width/2, floor);
                break;
            case RESHAPE:
                g.drawString("Reshape mode", d.width/2, floor);
                break;
        }
        painting = false;
    }
    
    
    /**
     *  CONTROL Section
     */
    
    /**
     * respond to custom button press
     * @param code the number of the button pressed (0 ... (Labels.length-1) )
     */
    public void doButton(int code) {
        
        switch(code) {
            case 0:  //  Button A
//                System.out.println("Reset");
                reset();
                break;
            case 1:  // Button B
//                System.out.println("Pan");
                mode = PAN;
                break;
            case 2:  // Button C
//                System.out.println("Zoom");
                mode = ZOOM;
                break;
            case 3:  // Button D
//                System.out.println("Rotate");
                mode = ROTATE;
                break;
            case 4:  // Button E
                Global.debug.println("Reshape");
                System.out.println("Reshape active");
                mode = RESHAPE;
                break;
        }
    }
    
    /**
     * Mouse clicked
     * @param e 
     */
    public void clicked(MouseEvent e) {
        System.out.println("Mouse Clicked");
    };
    /**
     * Mouse entered
     * @param e 
     */
    public void entered(MouseEvent e) {
    };
    /**
     * Mouse exited
     * @param e 
     */
    public void exited(MouseEvent e) {
    };
    /**
     * Mouse released
     * @param e 
     */
    public void released(MouseEvent e) {
//        System.out.println("Mouse Released");
        if(mode == RESHAPE) {
            Global.ctrlPointSelectedFlag = false;
        }
    };
    /**
     * Mouse pressed
     * @param e 
     */
    public void pressed(MouseEvent e) {
        baseX = e.getX();
        baseY = e.getY();
        String str = "";
        switch(mode) {
            case NONE:
                break;
            case PAN:
//                str = "Initial pan: " + Global.offset;
                baseOffset = new Vector(Global.offset);
                break;
            case ZOOM:
//                str = "Initial zoom: " + Global.scale;
                baseScale = Global.scale;
                break;
            case ROTATE:
                baseTheta = Global.theta;
                basePhi = Global.phi;
//                str = "base angles: theta = " + Global.theta + "; phi = " + Global.phi;
                break;
            case RESHAPE:
		pointedCoordinate = new Vector(baseX, baseY);
		Global.selectedCtrlPointIndex = thisPatch.getTheNearestCtrlPointIndex(pointedCoordinate);
		Global.ctrlPointSelectedFlag = Global.selectedCtrlPointIndex.k != -1;
                if(Global.ctrlPointSelectedFlag) {
                    baseCtl = new Vector(thisPatch.ctrlPoint[Global.selectedCtrlPointIndex.k][Global.selectedCtrlPointIndex.l]);
                } else baseCtl = null;
//                Global.debug.println("picked at " + pointedCoordinate + "; found " + Global.selectedCtrlPointIndex);
                break;
        }
//        if(mode != NONE) {
//            Global.debug.println(str);
//            System.out.println(str);            
//        }
    };
    /**
     * Mouse dragged
     * @param e 
     */
    public void dragged(MouseEvent e) {
        int x = e.getX();
        int y = e.getY();
        String str = "";
        double dx = ((double) x - baseX) / Global.screenSize.width;
        double dy = ((double) y - baseY) / Global.screenSize.height;
        switch(mode) {
            case NONE:
                break;
            case PAN:
                Global.offset = baseOffset.add(new Vector(x - baseX, y - baseY));
//                str = "New pan: " + Global.offset;
                break;
            case ZOOM:
                double val = ((0.25 * dy - 0.75) * dy + 1);
                Global.scale = baseScale * val * val;
//                str = "New zoom: " + Global.scale;
                break;
            case ROTATE:
                Global.theta = baseTheta + dx * Math.PI/2;
                Global.phi = basePhi - dy * Math.PI/2;
//                str = "new angles: theta = " + Global.theta + "; phi = " + Global.phi;
                break;
            case RESHAPE:
                if(Global.ctrlPointSelectedFlag) {
                    int i = Global.selectedCtrlPointIndex.k;
                    int j = Global.selectedCtrlPointIndex.l;
//		    Global.debug.println("changed ctrlPoint[" + i + "][" + j + "]: " + Global.ctrlPoint[i][j]);
                    Vector dp = new Vector((x - baseX)*Global.f/2, (y - baseY)*Global.f/2);
                    Vector it = thisPatch.ctrlPoint[Global.selectedCtrlPointIndex.k][Global.selectedCtrlPointIndex.l];
                    Global.incrementalInversePerspectiveTransformation(dp, baseCtl, it);
                    thisPatch.getBezierPointsCoordinate();
                    thisPatch.getBezierSurfacesNormalVector();
//		    Global.debug.println("    to: ctrlPoint[" + i + "][" + j + "]: " + Global.ctrlPoint[i][j]);
                }
                break;
        }
//        if(mode != NONE) {
//            Global.debug.println(str);
//            System.out.println(str);            
//        }
    };
    /**
     * Mouse moved
     * @param e 
     */
    public void moved(MouseEvent e) {
//        System.out.println("Mouse Moved");
    };
}